==> fwobac-1.6.tar.gz: fwobac-1.6/utf8_utils.c <==
/* $Id: utf8_utils.c,v 1.7 2018/06/12 05:29:18 rlr Exp $ */

/*
 * Copyright (c) 2018 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <ctype.h>
#include <string.h>

#include "utf8_utils.h"


/* UTF-8 encoding of UTF8_REPLACEMENT_CP... */
#define REPL      "\357\277\275"
/* ...and its length. */
#define REPL_LEN  3


bool
utf8_is_valid_str(const char *s)
{
	size_t	 encsz, n;
	uint32_t	 cp;
	bool	 v;

	n = 0;
	v = true;

	while (s[n] != '\0' && v) {
		/* Optimise the ASCII case. */
		if (isascii((unsigned char)s[n])) {
			/* Single byte. */
			n++;
		} else {
			utf8_strp_to_codept(&s[n], &cp, &encsz);
			if (cp == UTF8_INVALID_CP) {
				v = false;
			} else {
				n += encsz;
			}
		}
	}

	return v;
}


void
utf8_codept_to_strl(uint32_t cp, char *dst, size_t dstsz, size_t *encsz)
{
	size_t	 i, n;
	char	 c;

	if (cp <= 0x0000007FU) {
		c = cp & 0x0000007FU;
		n = 1;
	} else if (cp <= 0x0000077FU) {
		c = 0x000000C0U | (cp >> 6);
		n = 2;
	} else if (cp <= 0x0000FFFFU) {
		c = 0x000000E0U | (cp >> 12);
		n = 3;
	} else if (cp <= 0x0010FFFFU) {
		c = 0x000000F0U | (cp >> 18);
		n = 4;
	} else {
		n = 0;
	}

	if (n == 0 || n >= dstsz) {
		/* Invalid code point or insufficient space. */
		if (dstsz > 0) {
			dst[0] = '\0';
		}
	} else {
		dst[0] = c;
		for (i = 1; i < n; i++) {
			dst[i] = 0x00000080U
			    | ((cp >> ((n - i - 1) * 6)) & 0x0000003FU);
		}
		dst[n] = '\0';
	}
	*encsz = n;
}


void
utf8_strp_to_codept(const char *s, uint32_t *cp, size_t *encsz)
{
	size_t	 addtnl_bytes_xpctd, i, num_bytes;
	uint32_t	 codept;

	if (s[0] == '\0') {
		/* Empty string is treated as an error. */
		num_bytes = 0;
		addtnl_bytes_xpctd = 0;
		codept = UTF8_INVALID_CP;
	} else if (isascii((unsigned char)s[0])) {
		/* Encoded as a single byte. */
		num_bytes = 1;
		addtnl_bytes_xpctd = 0;
		codept = (unsigned char)s[0];
	} else if (((unsigned char)s[0] & 0xE0U) == 0xC0U) {
		/* Should be (and require) two bytes. */
		if (((unsigned char)s[1] & 0xC0U) == 0x80U
		    && ((unsigned char)s[0] & 0x1EU)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 1;
			codept = ((unsigned char)s[0] & 0x1FU);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else if (((unsigned char)s[0] & 0xF0U) == 0xE0U) {
		/* Should be (and require) three bytes. */
		if (((unsigned char)s[0] & 0x0FU)
		    || ((unsigned char)s[1] & 0x20U)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 2;
			codept = ((unsigned char)s[0] & 0x0FU);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else if (((unsigned char)s[0] & 0xF8U) == 0xF0U) {
		/* Should be (and require) four bytes. */
		if (((unsigned char)s[0] & 0x07U)
		    || ((unsigned char)s[1] & 0x30U)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 3;
			codept = ((unsigned char)s[0] & 0x07U);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else {
		/* Invalid byte. */
		num_bytes = 1;
		addtnl_bytes_xpctd = 0;
		codept = UTF8_INVALID_CP;
	}

	if (codept != UTF8_INVALID_CP) {
		/* Process any remaining bytes of encoding. */
		i = 1;
		while (i <= addtnl_bytes_xpctd
		    && ((unsigned char)s[i] & 0xC0U) == 0x80U) {
			codept = (codept << 6) | ((unsigned char)s[i] & 0x3FU);
			i++;
		}
		if (i <= addtnl_bytes_xpctd) {
			/* Invalid byte was encountered. */
			codept = UTF8_INVALID_CP;
			num_bytes = 1;  /* Blame the first byte. */
		} else {
			num_bytes = i;
		}
	}

	/* Check for code point out-of-range. */
	if (codept != UTF8_INVALID_CP && codept > 0x0010FFFFU) {
		/* Leave num_bytes untouched as the encoding itself was OK. */
		codept = UTF8_INVALID_CP;
	}

	*cp = codept;
	*encsz = num_bytes;
}


size_t
utf8_strlcpy(char *dst, const char *src, size_t dstsz)
{
	size_t	 d, s, z;
	uint32_t	 cp;
	bool	 elided;

	/*
	 * Valid multi-byte encodings are treated atomically.  Where there is
	 * insufficient space in the destination buffer to append the complete
	 * sequence, we cease all appending (other than a terminating NUL).
	 */
	elided = false;

	s = 0;
	d = 0;

	while (src[s] != '\0') {
		if (isascii((unsigned char)src[s])) {
			if (d + 1 < dstsz && !elided) {
				dst[d++] = src[s];
			}
			s++;
		} else {
			utf8_strp_to_codept(&src[s], &cp, &z);

			if (!elided) {
				if (d + (cp != UTF8_INVALID_CP ? z : REPL_LEN)
				    >= dstsz) {
					elided = true;
				} else {
					strlcpy(&dst[d], cp != UTF8_INVALID_CP
					    ? &src[s] : REPL, dstsz - d);
					d += (cp != UTF8_INVALID_CP ? z
					    : REPL_LEN);
				}
			}

			s += z;
		}
	}

	if (dstsz == 0) {
		/* We cannot NUL-terminate. */
	} else if (d < dstsz) {
		/* dst string is not truncated. */
		dst[d] = '\0';
	} else {
		/* NUL-terminate the truncated dst string. */
		dst[dstsz - 1] = '\0';
	}

	return d;
}

==> fwobac-2.1.tar.gz: fwobac-2.1/utf8_utils.c <==
/* $Id: utf8_utils.c,v 1.9 2020/02/06 23:36:23 rlr Exp $ */

/*
 * Copyright (c) 2018, 2020 Ross L Richardson <rueda@rlr.id.au>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <ctype.h>
#include <string.h>

#include "utf8_utils.h"


/* UTF-8 encoding of UTF8_REPLACEMENT_CP... */
#define REPL      "\357\277\275"
/* ...and its length. */
#define REPL_LEN  3


bool
utf8_is_valid_str(const char *s)
{
	size_t	 encsz, n;
	uint32_t	 cp;
	bool	 v;

	n = 0;
	v = true;

	while (s[n] != '\0' && v) {
		/* Optimise the ASCII case. */
		if (isascii((unsigned char)s[n])) {
			/* Single byte. */
			n++;
		} else {
			utf8_strp_to_codept(&s[n], &cp, &encsz);
			if (cp == UTF8_INVALID_CP) {
				v = false;
			} else {
				n += encsz;
			}
		}
	}

	return v;
}


void
utf8_codept_to_strl(uint32_t cp, char *dst, size_t dstsz, size_t *encsz)
{
	size_t	 n;
	char	 c;

	if (cp <= 0x0000007FU) {
		c = cp & 0x0000007FU;
		n = 1;
	} else if (cp <= 0x0000077FU) {
		c = 0x000000C0U | (cp >> 6);
		n = 2;
	} else if (cp <= 0x0000FFFFU) {
		c = 0x000000E0U | (cp >> 12);
		n = 3;
	} else if (cp <= 0x0010FFFFU) {
		c = 0x000000F0U | (cp >> 18);
		n = 4;
	} else {
		c = '\0';  /* Avoid spurious cppcheck error report. */
		n = 0;
	}

	if (n == 0 || n >= dstsz) {
		/* Invalid code point or insufficient space. */
		if (dstsz > 0) {
			dst[0] = '\0';
		}
	} else {
		size_t	 i;

		dst[0] = c;
		for (i = 1; i < n; i++) {
			dst[i] = 0x00000080U
			    | ((cp >> ((n - i - 1) * 6)) & 0x0000003FU);
		}
		dst[n] = '\0';
	}
	*encsz = n;
}


void
utf8_strp_to_codept(const char *s, uint32_t *cp, size_t *encsz)
{
	size_t	 addtnl_bytes_xpctd, num_bytes;
	uint32_t	 codept;

	if (s[0] == '\0') {
		/* Empty string is treated as an error. */
		num_bytes = 0;
		addtnl_bytes_xpctd = 0;
		codept = UTF8_INVALID_CP;
	} else if (isascii((unsigned char)s[0])) {
		/* Encoded as a single byte. */
		num_bytes = 1;
		addtnl_bytes_xpctd = 0;
		codept = (unsigned char)s[0];
	} else if (((unsigned char)s[0] & 0xE0U) == 0xC0U) {
		/* Should be (and require) two bytes. */
		if (((unsigned char)s[1] & 0xC0U) == 0x80U
		    && ((unsigned char)s[0] & 0x1EU)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 1;
			codept = ((unsigned char)s[0] & 0x1FU);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else if (((unsigned char)s[0] & 0xF0U) == 0xE0U) {
		/* Should be (and require) three bytes. */
		if (((unsigned char)s[0] & 0x0FU)
		    || ((unsigned char)s[1] & 0x20U)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 2;
			codept = ((unsigned char)s[0] & 0x0FU);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else if (((unsigned char)s[0] & 0xF8U) == 0xF0U) {
		/* Should be (and require) four bytes. */
		if (((unsigned char)s[0] & 0x07U)
		    || ((unsigned char)s[1] & 0x30U)) {
			num_bytes = 1;
			addtnl_bytes_xpctd = 3;
			codept = ((unsigned char)s[0] & 0x07U);
		} else {
			num_bytes = 1;
			addtnl_bytes_xpctd = 0;
			codept = UTF8_INVALID_CP;
		}
	} else {
		/* Invalid byte. */
		num_bytes = 1;
		addtnl_bytes_xpctd = 0;
		codept = UTF8_INVALID_CP;
	}

	if (codept != UTF8_INVALID_CP) {
		size_t	 i;

		/* Process any remaining bytes of encoding. */
		i = 1;
		while (i <= addtnl_bytes_xpctd
		    && ((unsigned char)s[i] & 0xC0U) == 0x80U) {
			codept = (codept << 6) | ((unsigned char)s[i] & 0x3FU);
			i++;
		}
		if (i <= addtnl_bytes_xpctd) {
			/* Invalid byte was encountered. */
			codept = UTF8_INVALID_CP;
			num_bytes = 1;  /* Blame the first byte. */
		} else {
			num_bytes = i;
		}
	}

	/* Check for code point out-of-range. */
	if (codept != UTF8_INVALID_CP && codept > 0x0010FFFFU) {
		/* Leave num_bytes untouched as the encoding itself was OK. */
		codept = UTF8_INVALID_CP;
	}

	*cp = codept;
	*encsz = num_bytes;
}


size_t
utf8_strlcpy(char *dst, const char *src, size_t dstsz)
{
	size_t	 d, s, z;
	uint32_t	 cp;
	bool	 elided;

	/*
	 * Valid multi-byte encodings are treated atomically.  Where there is
	 * insufficient space in the destination buffer to append the complete
	 * sequence, we cease all appending (other than a terminating NUL).
	 */
	elided = false;

	s = 0;
	d = 0;

	while (src[s] != '\0') {
		if (isascii((unsigned char)src[s])) {
			if (d + 1 < dstsz && !elided) {
				dst[d++] = src[s];
			}
			s++;
		} else {
			utf8_strp_to_codept(&src[s], &cp, &z);

			if (!elided) {
				if (d + (cp != UTF8_INVALID_CP ? z : REPL_LEN)
				    >= dstsz) {
					elided = true;
				} else {
					strlcpy(&dst[d], cp != UTF8_INVALID_CP
					    ? &src[s] : REPL, dstsz - d);
					d += (cp != UTF8_INVALID_CP ? z
					    : REPL_LEN);
				}
			}

			s += z;
		}
	}

	if (dstsz == 0) {
		/* We cannot NUL-terminate. */
	} else if (d < dstsz) {
		/* dst string is not truncated. */
		dst[d] = '\0';
	} else {
		/* NUL-terminate the truncated dst string. */
		dst[dstsz - 1] = '\0';
	}

	return d;
}
